/*
 * alangop.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

int main(int argc, char* argv[]) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Stringology algorithm access binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("unionEpsilon");
		allowed.push_back("unionCartesian");
		allowed.push_back("concatenationEpsilon");
		allowed.push_back("concatenation");
		allowed.push_back("intersectionCartesian");
		allowed.push_back("iterationEpsilon");
		allowed.push_back("iteration");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",	"Execute algorithm", true,	"",	&allowedVals);
		cmd.add(algorithm);

		TCLAP::ValueArg<std::string> input1(	"i",	"input1",	"First automaton",	false,	"-",	"file");
		TCLAP::ValueArg<std::string> input2(	"j",	"input2",	"Second automaton",	false,	"-",	"file");
		cmd.add( input1 );
		cmd.add( input2 );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc,argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "input1", input1.getValue ( ) );
		environment.setBinding ( "input2", input2.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		cli::Parser parser ( cli::Lexer ( "execute <#input1 > $language1" ) );
		parser.parse ( )->run ( environment );

		if ( algorithm.getValue ( ) == "unionEpsilon" || algorithm.getValue ( ) == "unionCartesian" || algorithm.getValue ( ) == "concatenationEpsilon" || algorithm.getValue ( ) == "concatenation" || algorithm.getValue ( ) == "intersectionCartesian") {
			parser = cli::Parser ( cli::Lexer ( "execute <#input2 > $language2" ) );
			parser.parse ( )->run ( environment );
		}

		measurements::end();
		measurements::start("Algorithm", measurements::Type::MAIN);

		if( algorithm.getValue() == "unionEpsilon") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomataUnionEpsilonTransition $language1 $language2 > $output" ) );
		} else if( algorithm.getValue() == "unionCartesian") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomataUnionCartesianProduct $language1 $language2 > $output" ) );
		} else if( algorithm.getValue() == "concatenationEpsilon") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomataConcatenationEpsilonTransition $language1 $language2 > $output" ) );
		} else if( algorithm.getValue() == "concatenation") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomataConcatenation $language1 $language2 > $output" ) );
		} else if( algorithm.getValue() == "intersectionCartesian") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomataIntersectionCartesianProduct $language1 $language2 > $output" ) );
		} else if( algorithm.getValue() == "iteration") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomatonIteration $language1 > $output" ) );
		} else if( algorithm.getValue() == "iterationEpsilon") {
			parser = cli::Parser ( cli::Lexer ( "execute automaton::transform::AutomatonIterationEpsilonTransition $language1 > $output" ) );
		} else {
			throw exception::CommonException( "Invalid algorithm" );
		}
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::start ( "Output write", measurements::Type::AUXILIARY );

		parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::end();

		if(measure.getValue()) common::Streams::measure << measurements::results() << std::endl;

		return 0;
	} catch( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
