/*
 * EpsilonClosure.h
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef EPSILON_CLOSURE_H_
#define EPSILON_CLOSURE_H_

#include <algorithm>
#include <queue>
#include <set>
#include <map>

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <regexp/properties/RegExpEpsilon.h>

namespace automaton {

namespace properties {

class EpsilonClosure {
public:
	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	template < class SymbolType, class EpsilonType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm, const StateType & state );
	template < class SymbolType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm, const StateType & state );
	template < class SymbolType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::NFA < SymbolType, StateType > & fsm, const StateType & state );
	template < class SymbolType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::DFA < SymbolType, StateType > & fsm, const StateType & state );
	template < class SymbolType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::ExtendedNFA < SymbolType, StateType > & fsm, const StateType & state );
	template < class SymbolType, class StateType >
	static ext::set<StateType> epsilonClosure( const automaton::CompactNFA < SymbolType, StateType > & fsm, const StateType & state );
};

template < class SymbolType, class EpsilonType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	std::queue<StateType> queue;
	ext::map<StateType, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) ) {
		StateType p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		auto tos = fsm.getTransitions( ).find(ext::make_pair ( p, ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( ) ) );
		if(tos == fsm.getTransitions().end()) continue;

		for( const auto & to : tos->second )
			if( visited [ to ] == false )
				queue.push( to );
	}

	return closure;
}

template < class SymbolType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	closure.insert(q);
	return closure;
}

template < class SymbolType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::NFA < SymbolType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	closure.insert(q);
	return closure;
}

template < class SymbolType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::DFA < SymbolType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	closure.insert(q);
	return closure;
}

template < class SymbolType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::ExtendedNFA < SymbolType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	std::queue<StateType> queue;
	ext::map<StateType, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) ) {
		StateType p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

template < class SymbolType, class StateType >
ext::set<StateType> EpsilonClosure::epsilonClosure( const automaton::CompactNFA < SymbolType, StateType > & fsm, const StateType & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	ext::set<StateType> closure;
	std::queue<StateType> queue;
	ext::map<StateType, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) ) {
		StateType p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( transition.first.second.size() == 0 )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

} /* namespace properties */

} /* namespace automaton */

#endif /* EPSILON_CLOSURE_H_ */
