/*
 * HopcroftGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 */

#ifndef GRAPH_HOPCROFT_GRAPH_H_
#define GRAPH_HOPCROFT_GRAPH_H_

#include <vector>

#include <graph/GraphClasses.h>

namespace graph
{

namespace isomorphism
{

namespace hopcroft
{

class Vertex;
class EdgeEnd;
class Face;

enum Direction {
	ClockWise,
	CounterClockWise
};

class Graph
{
public:
	explicit Graph(const UndirectedGraph &graph);
	~Graph();

	// Reduction-specific modifications
	EdgeEnd *addInnerEdge(EdgeEnd *e1, EdgeEnd *e2);
	void removeDistinguishedEdgeAndVertex(EdgeEnd *e);
	void addAndReconnectVertexIntoTriangularFace(EdgeEnd *e);
	void removeFourDegreeExceptionVertex(Vertex *v, EdgeEnd *e);

	void removeEdge(EdgeEnd *e, bool removeVertices = true);

	void removeFace(Face *f);
	void removeVertex(Vertex *v);
	void removeEdgeEnd(EdgeEnd *e);

	ext::vector<ext::vector<EdgeEnd*>> findClumps(Vertex *v, Vertex *w) const;

	void dump();
	void dumpLabels();

	ext::vector<Face*> faces;
	ext::vector<Vertex*> vertices;
	ext::vector<EdgeEnd*> edge_ends;

private:
	Graph(const Graph &other) = delete;

	void init(const UndirectedGraph &graph);
};

class Vertex
{
public:
	explicit Vertex();

	bool isKnot() const;
	bool isCorolla() const;

	bool isStar() const;
	bool isDumbell() const;
	bool isLoopVertex() const;
	bool isSpokeCenter() const;

	bool hasNonLoopEdge() const;
	bool hasNonSpokeEdge() const;
	bool isDegreeIsolated() const;

	ext::vector<Face*> neighborFaces(Direction d = ClockWise) const;
	ext::vector<EdgeEnd*> neighborEdges(Direction d = ClockWise) const;
	void forEachEdge(std::function<bool(EdgeEnd*)> cb, Direction d = ClockWise) const;

	void updateDegree();

	int vlabel;
	int degree;
	EdgeEnd *vedge;

	// Debug
	std::string dump() const;
	std::string out;
};

class EdgeEnd
{
public:
	explicit EdgeEnd();

	bool isSpoke() const;
	bool isLoopEdge() const;

	bool isMultiEdge() const;

	int elabel;
	Vertex *vertex;
	Face *ccwface;
	EdgeEnd *ccwedge;
	EdgeEnd *cwedge;
	EdgeEnd *otheredge; // other edge-end

	// Debug
	std::string dump() const;
};

class Face
{
public:
	explicit Face();

	int edges;
	EdgeEnd *fedge;
};

} // namespace hopcroft

} // namespace isomorphism

} // namespace graph

#endif // GRAPH_HOPCROFT_GRAPH_H_
