/*
 * NonlinearFullAndLinearIndex.h
 *
 *  Created on: Apr 13, 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_
#define ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_

#include <string>
#include <sstream>

#include <alphabet/RankedSymbol.h>
#include <alphabet/xml/RankedSymbol.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsVector.h>
#include <container/xml/ObjectsVector.h>

#include <primitive/xml/Integer.h>

#include <indexes/stringology/PositionHeap.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType, class RankType = DefaultRankType >
class NonlinearFullAndLinearIndex final : public alib::ObjectBase {
protected:
	indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > m_StringIndex;
	ext::vector < int > m_JumpTable;
	ext::vector < unsigned > m_Repeats;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit NonlinearFullAndLinearIndex ( indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > stringIndex, ext::vector < int > jumpTable, ext::vector < unsigned > repeats );

	/**
	 * @return Root node of the trie
	 */
	const indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > & getStringIndex ( ) const &;

	indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > && getStringIndex ( ) &&;

	/**
	 * @return subtree jump table
	 */
	const ext::vector < int > & getJumps ( ) const &;

	ext::vector < int > && getJumps ( ) &&;

	/**
	 * @return repeats
	 */
	const ext::vector < unsigned > & getRepeats ( ) const &;

	ext::vector < unsigned > && getRepeats ( ) &&;

	const ext::vector < common::ranked_symbol < SymbolType, RankType > > & getString ( ) const &;

	ext::vector < common::ranked_symbol < SymbolType, RankType > > && getString ( ) &&;

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return m_StringIndex.getAlphabet ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( m_StringIndex.getAlphabet ( ) );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setStringIndex ( indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > stringIndex );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return m_StringIndex.removeSymbolFromAlphabet ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NonlinearFullAndLinearIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef NonlinearFullAndLinearIndex < > normalized_type;
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType, class RankType >
NonlinearFullAndLinearIndex < SymbolType, RankType >::NonlinearFullAndLinearIndex ( indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > stringIndex, ext::vector < int > jumpTable, ext::vector < unsigned > repeats ) : m_StringIndex ( std::move ( stringIndex ) ), m_JumpTable ( std::move ( jumpTable ) ), m_Repeats ( std::move ( repeats ) ) {
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType >::clone ( ) const {
	return new NonlinearFullAndLinearIndex ( * this );
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType >::plunder ( ) && {
	return new NonlinearFullAndLinearIndex ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > & NonlinearFullAndLinearIndex < SymbolType, RankType >::getStringIndex ( ) const & {
	return m_StringIndex;
}

template < class SymbolType, class RankType >
indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > && NonlinearFullAndLinearIndex < SymbolType, RankType >::getStringIndex ( ) && {
	return std::move ( m_StringIndex );
}

template < class SymbolType, class RankType >
const ext::vector < int > & NonlinearFullAndLinearIndex < SymbolType, RankType >::getJumps ( ) const & {
	return m_JumpTable;
}

template < class SymbolType, class RankType >
ext::vector < int > && NonlinearFullAndLinearIndex < SymbolType, RankType >::getJumps ( ) && {
	return std::move ( m_JumpTable );
}

template < class SymbolType, class RankType >
const ext::vector < unsigned > & NonlinearFullAndLinearIndex < SymbolType, RankType >::getRepeats ( ) const & {
	return m_Repeats;
}

template < class SymbolType, class RankType >
ext::vector < unsigned > && NonlinearFullAndLinearIndex < SymbolType, RankType >::getRepeats ( ) && {
	return std::move ( m_Repeats );
}

template < class SymbolType, class RankType >
const ext::vector < common::ranked_symbol < SymbolType, RankType > > & NonlinearFullAndLinearIndex < SymbolType, RankType >::getString ( ) const & {
	return m_StringIndex.getString ( );
}

template < class SymbolType, class RankType >
ext::vector < common::ranked_symbol < SymbolType, RankType > > && NonlinearFullAndLinearIndex < SymbolType, RankType >::getString ( ) && {
	return std::move ( m_StringIndex ).getString ( );
}

template < class SymbolType, class RankType >
void NonlinearFullAndLinearIndex < SymbolType, RankType >::setStringIndex ( indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > stringIndex ) {
	this->m_StringIndex = std::move ( stringIndex );
}

template < class SymbolType, class RankType >
void NonlinearFullAndLinearIndex < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(NonlinearFullAndLinearIndex " << this->m_StringIndex << ", " << m_JumpTable << ")";
}

template < class SymbolType, class RankType >
int NonlinearFullAndLinearIndex < SymbolType, RankType >::compare ( const NonlinearFullAndLinearIndex & other ) const {
	auto first = ext::tie ( getStringIndex ( ), getJumps ( ), getRepeats ( ) );
	auto second = ext::tie ( other.getStringIndex ( ), other.getJumps ( ), other.getRepeats ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
NonlinearFullAndLinearIndex < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType, RankType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

namespace alib {

template < class SymbolType, class RankType >
struct normalize < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType >, typename std::enable_if < ! std::is_same < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType >, indexes::arbology::NonlinearFullAndLinearIndex < > >::value >::type > {
	static indexes::arbology::NonlinearFullAndLinearIndex < > eval ( indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > && value ) {
		indexes::stringology::PositionHeap < common::ranked_symbol < > > stringIndex = alib::normalize < indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > >::eval ( std::move ( value ).getStringIndex ( ) );

		return indexes::arbology::NonlinearFullAndLinearIndex < > ( std::move ( stringIndex ), std::move ( value ).getJumps ( ), std::move ( value ).getRepeats ( ) );
	}
};

template < class SymbolType, class RankType >
struct xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > > {
	static indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > & data );
};

template < class SymbolType, class RankType >
indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > stringIndex = alib::xmlApi < indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > >::parse ( input );
	ext::vector < int > jumps = alib::xmlApi < ext::vector < int > >::parse ( input );
	ext::vector < unsigned > repeats = alib::xmlApi < ext::vector < unsigned > >::parse ( input );

	indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > res ( std::move ( stringIndex ), std::move ( jumps ), std::move ( repeats ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType, class RankType >
bool xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType, class RankType >
const std::string & xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > >::xmlTagName ( ) {
	static std::string xmlTagName = "NonlinearFullAndLinearIndex";

	return xmlTagName;
}

template < class SymbolType, class RankType >
void xmlApi < indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > >::compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearFullAndLinearIndex < SymbolType, RankType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < indexes::stringology::PositionHeap < common::ranked_symbol < SymbolType, RankType > > >::compose ( output, index.getStringIndex ( ) );
	alib::xmlApi < ext::vector < int > >::compose ( output, index.getJumps ( ) );
	alib::xmlApi < ext::vector < unsigned > >::compose ( output, index.getRepeats ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace alib */

#endif /* ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_ */
