/*
 * UnboundedRegExpSymbol.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef UNBOUNDED_REG_EXP_SYMBOL_H_
#define UNBOUNDED_REG_EXP_SYMBOL_H_

#include <sstream>

#include "UnboundedRegExpElement.h"
#include <alphabet/common/SymbolNormalize.h>

namespace regexp {

/**
 * Represents symbol in the regular expression. Contains name of the symbol.
 */
template < class SymbolType >
class UnboundedRegExpSymbol : public UnboundedRegExpElement < SymbolType >, public ext::NullaryNode < ext::smart_ptr < UnboundedRegExpElement < SymbolType > >, ext::smart_ptr < const UnboundedRegExpElement < SymbolType > >, UnboundedRegExpSymbol < SymbolType > > {
	SymbolType m_symbol;

public:
	void accept ( typename UnboundedRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit UnboundedRegExpSymbol ( SymbolType symbol );

	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	UnboundedRegExpSymbol < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	UnboundedRegExpSymbol < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc UnboundedRegExpElement::cloneAsFormal() const
	 */
	ext::smart_ptr < FormalRegExpElement < SymbolType > > asFormal ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( ext::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const ext::set < SymbolType > & alphabet ) const override;

	int compare ( const UnboundedRegExpElement < SymbolType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	int compare ( const UnboundedRegExpSymbol & ) const;

	/**
	 * @copydoc UnboundedRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	/**
	 * Returns the string representation of RegExp Symbol.
	 */
	const SymbolType & getSymbol ( ) const;

	explicit operator std::string ( ) const override;

	virtual ext::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		return ext::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > ( new UnboundedRegExpSymbol < DefaultSymbolType > ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( m_symbol ) ) ) );
	}
};

} /* namespace regexp */

#include "UnboundedRegExpSymbol.h"
#include "../formal/FormalRegExpSymbol.h"
#include <sstream>

namespace regexp {

template < class SymbolType >
UnboundedRegExpSymbol < SymbolType >::UnboundedRegExpSymbol ( SymbolType symbol ) : m_symbol ( std::move ( symbol ) ) {
}

template < class SymbolType >
UnboundedRegExpSymbol < SymbolType > * UnboundedRegExpSymbol < SymbolType >::clone ( ) const {
	return new UnboundedRegExpSymbol ( * this );
}

template < class SymbolType >
UnboundedRegExpSymbol < SymbolType > * UnboundedRegExpSymbol < SymbolType >::plunder ( ) && {
	return new UnboundedRegExpSymbol ( std::move ( * this ) );
}

template < class SymbolType >
ext::smart_ptr < FormalRegExpElement < SymbolType > > UnboundedRegExpSymbol < SymbolType >::asFormal ( ) const {
	return ext::smart_ptr < FormalRegExpElement < SymbolType > > ( new FormalRegExpSymbol < SymbolType > ( this->m_symbol ) );
}

template < class SymbolType >
int UnboundedRegExpSymbol < SymbolType >::compare ( const UnboundedRegExpSymbol & other ) const {
	static ext::compare < decltype ( m_symbol ) > comp;

	return comp ( m_symbol, other.m_symbol );
}

template < class SymbolType >
void UnboundedRegExpSymbol < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExpSymbol " << m_symbol << ")";
}

template < class SymbolType >
bool UnboundedRegExpSymbol < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	return symbol == this->m_symbol;
}

template < class SymbolType >
bool UnboundedRegExpSymbol < SymbolType >::checkAlphabet ( const ext::set < SymbolType > & alphabet ) const {
	return alphabet.count ( m_symbol );
}

template < class SymbolType >
void UnboundedRegExpSymbol < SymbolType >::computeMinimalAlphabet ( ext::set < SymbolType > & alphabet ) const {
	alphabet.insert ( this->m_symbol );
}

template < class SymbolType >
const SymbolType & UnboundedRegExpSymbol < SymbolType >::getSymbol ( ) const {
	return this->m_symbol;
}

template < class SymbolType >
UnboundedRegExpSymbol < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_SYMBOL_H_ */
