/*
 * FormalRTEAlternation.h
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_ALTERNATION_H_
#define FORMAL_RTE_ALTERNATION_H_

#include <sstream>
#include <utility>

#include "FormalRTEElement.h"

#include <exception/CommonException.h>

namespace rte {

/**
 * Represents alternation operator in the regular expression. Contains list of FormalRTEElement
 * as operands of the operator.
 */
template < class SymbolType, class RankType >
class FormalRTEAlternation : public FormalRTEElement < SymbolType, RankType >, public ext::BinaryNode < ext::smart_ptr < FormalRTEElement < SymbolType, RankType > >, ext::smart_ptr < FormalRTEElement < SymbolType, RankType > >, FormalRTEAlternation < SymbolType, RankType > > {
public:
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit FormalRTEAlternation ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right );
	explicit FormalRTEAlternation ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right );

	/**
	 * @copydoc FormalRTEElement::clone() const
	 */
	virtual FormalRTEAlternation < SymbolType, RankType > * clone ( ) const override;

	/**
	 * @copydoc FormalRTEElement::plunder() const
	 */
	virtual FormalRTEAlternation < SymbolType, RankType > * plunder ( ) && override;

	/**
	 * @copydoc FormalRTEElement::testSymbol() const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement::computeMinimalAlphabet()
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement::checkAlphabet()
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @return elements
	 */
	const FormalRTEElement < SymbolType, RankType > & getLeftElement ( ) const;
	const FormalRTEElement < SymbolType, RankType > & getRightElement ( ) const;

	/**
	 * @return elements
	 */
	FormalRTEElement < SymbolType, RankType > & getLeftElement ( );
	FormalRTEElement < SymbolType, RankType > & getRightElement ( );

	/**
	 * @param element to append
	 */
	void setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element );
	void setLeftElement ( FormalRTEElement < SymbolType, RankType > && element );

	/**
	 * @param element to append
	 */
	void setRightElement ( const FormalRTEElement < SymbolType, RankType > & element );
	void setRightElement ( FormalRTEElement < SymbolType, RankType > && element );

	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	int compare ( const FormalRTEAlternation & ) const;

	/**
	 * @copydoc FormalRTEElement::operator>>() const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	virtual explicit operator std::string ( ) const override;

	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( new FormalRTEAlternation < DefaultSymbolType, DefaultRankType > ( std::move ( * std::move ( getLeftElement ( ) ).normalize ( ) ), std::move ( * std::move ( getRightElement ( ) ).normalize ( ) ) ) );
	}
};

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::FormalRTEAlternation ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right ) : ext::BinaryNode < ext::smart_ptr < FormalRTEElement < SymbolType, RankType > >, ext::smart_ptr < FormalRTEElement < SymbolType, RankType > >, FormalRTEAlternation < SymbolType, RankType > > ( ext::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( left ).plunder ( ) ), ext::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( right ).plunder ( ) ) ) {
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::FormalRTEAlternation ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right ) : FormalRTEAlternation ( ext::move_copy ( left ), ext::move_copy ( right ) ) {
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getLeftElement ( ) const {
	return * this->getLeft ( );
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getRightElement ( ) const {
	return * this->getRight ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getLeftElement ( ) {
	return * this->getLeft ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getRightElement ( ) {
	return * this->getRight ( );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setLeftElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setLeftElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setLeft ( ext::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setRightElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setRightElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setRightElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setRight ( ext::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType > * FormalRTEAlternation < SymbolType, RankType >::clone ( ) const {
	return new FormalRTEAlternation ( * this );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType > * FormalRTEAlternation < SymbolType, RankType >::plunder ( ) && {
	return new FormalRTEAlternation ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTEAlternation < SymbolType, RankType >::compare ( const FormalRTEAlternation & other ) const {
	int res = getLeftElement ( ).compare ( other.getLeftElement ( ) );

	if ( res == 0 ) res = getRightElement ( ).compare ( other.getRightElement ( ) );

	return res;
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTEAlternation";
	out << " " << getLeftElement ( );
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType, class RankType >
bool FormalRTEAlternation < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const {
	if ( getLeftElement ( ).testSymbol ( symbol ) ) return true;

	if ( getRightElement ( ).testSymbol ( symbol ) ) return true;

	return false;
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	getLeftElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
	getRightElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTEAlternation < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return getLeftElement ( ).checkAlphabet ( alphabetF, alphabetK ) && getRightElement ( ).checkAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_ALTERNATION_H_ */
