/*
 * UnrankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_TREE_H_
#define UNRANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "../TreeFeatures.h"
#include "../UnrankedTreeBase.h"
#include "../TreeException.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include <tree/common/TreeNormalize.h>

#include <core/normalize.hpp>

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType >
class UnrankedTree final : public UnrankedTreeBase, public alib::Components < UnrankedTree < SymbolType >, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
	ext::tree < SymbolType > m_content;

	void checkAlphabet ( const ext::tree < SymbolType > & pattern ) const;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedTree ( ext::set < SymbolType > alphabet, ext::tree < SymbolType > tree );
	explicit UnrankedTree ( ext::tree < SymbolType > tree );
	template < class RankType >
	explicit UnrankedTree ( const RankedTree < SymbolType, RankType > & other );

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( const ext::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return m_content of the tree
	 */
	const ext::tree < SymbolType > & getContent ( ) const &;

	ext::tree < SymbolType > && getContent ( ) &&;

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( ext::tree < SymbolType > data );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedTree";

		return xmlTagName;
	}

	static UnrankedTree parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const UnrankedTree & tree );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef UnrankedTree < > normalized_type;
};

} /* namespace tree */

#include "../ranked/RankedTree.h"

namespace tree {

template < class SymbolType >
UnrankedTree < SymbolType >::UnrankedTree ( ext::set < SymbolType > alphabet, ext::tree < SymbolType > tree ) : alib::Components < UnrankedTree, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ), m_content ( std::move ( tree ) ) {
	checkAlphabet ( m_content );
}

template < class SymbolType >
UnrankedTree < SymbolType >::UnrankedTree ( ext::tree < SymbolType > pattern ) : UnrankedTree ( ext::set < SymbolType > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ), pattern ) {
}

template < class SymbolType >
template < class RankType >
UnrankedTree < SymbolType >::UnrankedTree ( const RankedTree < SymbolType, RankType > & other ) : UnrankedTree ( TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

template < class SymbolType >
UnrankedTreeBase * UnrankedTree < SymbolType >::clone ( ) const {
	return new UnrankedTree ( * this );
}

template < class SymbolType >
UnrankedTreeBase * UnrankedTree < SymbolType >::plunder ( ) && {
	return new UnrankedTree ( std::move ( * this ) );
}

template < class SymbolType >
const ext::tree < SymbolType > & UnrankedTree < SymbolType >::getContent ( ) const & {
	return m_content;
}

template < class SymbolType >
ext::tree < SymbolType > && UnrankedTree < SymbolType >::getContent ( ) && {
	return std::move ( m_content );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::checkAlphabet ( const ext::tree < SymbolType > & data ) const {
	ext::set < SymbolType > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	ext::set < SymbolType > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::setTree ( ext::tree < SymbolType > data ) {
	checkAlphabet ( data );

	this->m_content = std::move ( data );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnrankedTree " << this->m_content << ")";
}

template < class SymbolType >
int UnrankedTree < SymbolType >::compare ( const UnrankedTree & other ) const {
	auto first = ext::tie ( m_content, getAlphabet() );
	auto second = ext::tie ( other.m_content, other.getAlphabet() );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::nicePrint ( std::ostream & os ) const {
	m_content.nicePrint ( os );
}

template < class SymbolType >
UnrankedTree < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
UnrankedTree < SymbolType > UnrankedTree < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedTree::getXmlTagName() );
	ext::set < SymbolType > rankedAlphabet = TreeFromXMLParser::parseAlphabet < SymbolType > ( input );
	ext::tree < SymbolType > root = TreeFromXMLParser::parseTreeContent < SymbolType > ( input );
	UnrankedTree < SymbolType > tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedTree::getXmlTagName() );
	return tree;
}

template < class SymbolType >
void UnrankedTree < SymbolType >::compose ( ext::deque < sax::Token > & out, const UnrankedTree & tree ) {
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, tree.getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, tree.getContent ( ) );
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* UnrankedTree < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace alib {

template < class SymbolType >
class ComponentConstraint< ::tree::UnrankedTree < SymbolType >, SymbolType, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedTree < SymbolType > & tree, const SymbolType & symbol ) {
		const ext::tree<SymbolType>& m_content = tree.getContent ( );
		return std::find(m_content.prefix_begin(), m_content.prefix_end(), symbol) != m_content.prefix_end();
	}

	static bool available ( const ::tree::UnrankedTree < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedTree < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < tree::UnrankedTree < SymbolType >, typename std::enable_if < ! std::is_same < tree::UnrankedTree < SymbolType >, tree::UnrankedTree < > >::value >::type > {
	static tree::UnrankedTree < > eval ( tree::UnrankedTree < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::tree < DefaultSymbolType > content = tree::TreeNormalize::normalizeTree ( std::move ( value ).getContent ( ) );

		return tree::UnrankedTree < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace alib */

#endif /* UNRANKED_TREE_H_ */
