/*
 * AdjacencyMatrixUndirectedGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#ifndef ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_
#define ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_

#include <set>
#include <vector>
#include <unordered_map>

#include "UndirectedGraph.h"

namespace graph {

class DirectedGraph;

/*
 *  AdjacencyMatrix representation
 *
 *  unordered_map:
 *    [ -- ][ n1 ][ n2 ][ n3 ][ n4 ]
 *    [ n1 ]  0     2     1     1
 *    [ n2 ]  2     0     0     1
 *    [ n3 ]  1     0     0     1
 *    [ n4 ]  1     1     1     0
 *
 *  edges:
 *    n1 <-> n2, n1 <-> n2, n1 <-> n3, n1 <-> n4
 *    n2 <-> n4
 *    n3 <-> n4
 *
 *  undirected edge n1 <-> n2 is represented as
 *  directed edge n1 -> n2 and directed edge n2 -> n1
 *
 */

class AdjacencyMatrixUndirectedGraph final : public UndirectedGraph {
	ext::unordered_map< Node, ext::unordered_map< Node, int > > adj;
	ext::set< UndirectedEdge > edges;

public:
	AdjacencyMatrixUndirectedGraph();

	/// constructs an undirected adjacency matrix from any representation of undirected graph
	AdjacencyMatrixUndirectedGraph( const UndirectedGraph & graph );
	/// converts directed graph into undirected by removing edge direction
	AdjacencyMatrixUndirectedGraph( const DirectedGraph & dgraph );

	virtual GraphBase * clone() const override;
	virtual GraphBase * plunder() && override;

	virtual REPRESENTATION getRepresentation() const override;
	virtual std::string className() const override;

	virtual ext::set< Node > getNodes() const override;
	virtual ext::set< UndirectedEdge > getEdges() const override;

	virtual size_t nodeCount() const override;
	virtual size_t edgeCount() const override;

	virtual bool hasNode( const Node & node ) const override;
	virtual bool hasEdge( const UndirectedEdge & edge ) const override;

	virtual bool hasEdge( const Node & first, const Node & second ) const override;
	virtual ext::set< UndirectedEdge > findEdges( const Node & first, const Node & second ) const override;

	virtual ext::set< Node > neighbors( const Node & node ) const override;
	virtual ext::set< UndirectedEdge > neighborEdges( const Node & node ) const override;

	virtual bool addNode( const Node & node ) override;
	virtual bool addNode( const Node & node, int value ) override;
	virtual bool removeNode( const Node & node ) override;

	virtual bool addEdge( const UndirectedEdge & edge ) override;
	virtual bool addEdge( const UndirectedEdge & edge, int value ) override;
	virtual bool removeEdge( const UndirectedEdge & edge ) override;

	static AdjacencyMatrixUndirectedGraph parse( ext::deque< sax::Token >::iterator & input );
	static void compose( ext::deque< sax::Token > & out, const AdjacencyMatrixUndirectedGraph & graph );

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "AdjacencyMatrixUndirectedGraph";

		return xmlTagName;
	}

	alib::ObjectBase * inc ( ) && override;

	typedef AdjacencyMatrixUndirectedGraph normalized_type;
};

} // namespace graph

#endif // ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_
