/*
 * TreeFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "TreeFromStringParser.h"
#include "exception/CommonException.h"
#include "tree/TreeClasses.h"
#include <tree/Tree.h>

#include "../StringApi.hpp"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/NonlinearVariableSymbol.h"

namespace tree {

Tree TreeFromStringParser::parseTree ( std::istream & input ) {
	return parseTree ( input, ext::set < FEATURES > ( { FEATURES::RANKED_TREE, FEATURES::RANKED_PATTERN, FEATURES::RANKED_NONLINEAR_PATTERN, FEATURES::UNRANKED_TREE, FEATURES::UNRANKED_PATTERN, FEATURES::UNRANKED_NONLINEAR_PATTERN } ) );
}

Tree TreeFromStringParser::parseTree ( std::istream & input, const ext::set < FEATURES > & features ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			return Tree ( UnrankedPattern < > ( alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( ), ext::tree < DefaultSymbolType > ( alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( ), { } ) ) );
		} else {
			return Tree ( RankedPattern < > ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType, primitive::Unsigned > > ( ), ext::tree < common::ranked_symbol < DefaultSymbolType > > ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType > > ( ), { } ) ) );
		}
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			DefaultSymbolType subtreeWildcard = alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( );
			DefaultSymbolType nonlinearVariable ( alphabet::NonlinearVariableSymbol < DefaultSymbolType > ( DefaultSymbolType ( token.value ) ) );
			return Tree ( UnrankedNonlinearPattern < > ( subtreeWildcard, { nonlinearVariable }, ext::tree < DefaultSymbolType > ( subtreeWildcard, { } ) ) );
		} else {
			common::ranked_symbol < DefaultSymbolType > subtreeWildcard = alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType > > ( );
			common::ranked_symbol < DefaultSymbolType > nonlinearVariable ( DefaultSymbolType ( alphabet::NonlinearVariableSymbol < DefaultSymbolType > ( DefaultSymbolType ( token.value ) ) ), 0 );
			return Tree ( RankedNonlinearPattern < > ( subtreeWildcard, { nonlinearVariable }, ext::tree < common::ranked_symbol < DefaultSymbolType > > ( nonlinearVariable, { } ) ) );
		}
	} else {
		TreeFromStringLexer::putback ( input, token );
		DefaultSymbolType symbol = alib::stringApi < DefaultSymbolType >::parse ( input );

		unsigned rank = 0;

		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK ) {
			ext::set < common::ranked_symbol < DefaultSymbolType > > nonlinearVariables;

			rank = ext::from_string < unsigned > ( token.value );

			ext::vector < ext::tree < common::ranked_symbol < DefaultSymbolType > > > childs;
			bool isPattern = false;

			for ( unsigned i = 0; i < rank; i++ )
				childs.emplace_back ( parseRankedContent ( input, isPattern, nonlinearVariables ) );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				RankedNonlinearPattern < > tree ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType > > ( ), std::move ( nonlinearVariables ), ext::tree < common::ranked_symbol < DefaultSymbolType > > ( common::ranked_symbol < DefaultSymbolType > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				RankedPattern < > tree ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType > > ( ), ext::tree < common::ranked_symbol < DefaultSymbolType > > ( common::ranked_symbol < DefaultSymbolType > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				RankedTree < > tree ( ext::tree < common::ranked_symbol < DefaultSymbolType > > ( common::ranked_symbol < DefaultSymbolType > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		} else {
			ext::set < DefaultSymbolType > nonlinearVariables;

			ext::vector < ext::tree < DefaultSymbolType > > childs;
			bool isPattern = false;

			while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
				TreeFromStringLexer::putback ( input, token );
				childs.emplace_back ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) );
				token = TreeFromStringLexer::next ( input );
			}

			if ( token.type != TreeFromStringLexer::TokenType::BAR )
				throw exception::CommonException ( "Missing bar" );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				UnrankedNonlinearPattern < > tree ( alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( ), std::move ( nonlinearVariables ), ext::tree < DefaultSymbolType > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				UnrankedPattern < > tree ( alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( ), ext::tree < DefaultSymbolType > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				UnrankedTree < > tree ( ext::tree < DefaultSymbolType > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		}

		throw exception::CommonException ( "Invalid input" );
	}
}

ext::tree < common::ranked_symbol < DefaultSymbolType > > TreeFromStringParser::parseRankedContent ( std::istream & input, bool & isPattern, ext::set < common::ranked_symbol < DefaultSymbolType > > & nonlinearVariables ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		isPattern = true;
		return ext::tree < common::ranked_symbol < DefaultSymbolType > > ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < DefaultSymbolType > > ( ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		isPattern = true;
		common::ranked_symbol < DefaultSymbolType > nonlinearVariable ( DefaultSymbolType ( alphabet::NonlinearVariableSymbol < DefaultSymbolType > ( DefaultSymbolType ( token.value ) ) ), 0 );
		nonlinearVariables.insert ( nonlinearVariable );
		return ext::tree < common::ranked_symbol < DefaultSymbolType > > ( std::move ( nonlinearVariable ), { } );
	} else {
		TreeFromStringLexer::putback ( input, token );
		DefaultSymbolType symbol = alib::stringApi < DefaultSymbolType >::parse ( input );

		unsigned rank = 0;
		ext::vector < ext::tree < common::ranked_symbol < DefaultSymbolType > > > childs;

		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK )
			rank = ext::from_string < unsigned > ( token.value );
		else
			throw exception::CommonException ( "Missing rank" );

		for ( unsigned i = 0; i < rank; i++ )
			childs.emplace_back ( ext::tree < common::ranked_symbol < DefaultSymbolType > > ( parseRankedContent ( input, isPattern, nonlinearVariables ) ) );

		return ext::tree < common::ranked_symbol < DefaultSymbolType > > ( common::ranked_symbol < DefaultSymbolType > ( std::move ( symbol ), rank ), std::move ( childs ) );
	}
}

ext::tree < DefaultSymbolType > TreeFromStringParser::parseUnrankedContent ( std::istream & input, bool & isPattern, ext::set < DefaultSymbolType > & nonlinearVariables ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		return ext::tree < DefaultSymbolType > ( alphabet::SubtreeWildcardSymbol::instance < DefaultSymbolType > ( ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		DefaultSymbolType nonlinearVariable ( alphabet::NonlinearVariableSymbol < DefaultSymbolType > ( DefaultSymbolType ( token.value ) ) );
		nonlinearVariables.insert ( nonlinearVariable );
		return ext::tree < DefaultSymbolType > ( std::move ( nonlinearVariable ), { } );
	} else {
		TreeFromStringLexer::putback ( input, token );
		DefaultSymbolType symbol = alib::stringApi < DefaultSymbolType >::parse ( input );

		ext::vector < ext::tree < DefaultSymbolType > > childs;

		token = TreeFromStringLexer::next ( input );

		while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
			TreeFromStringLexer::putback ( input, token );
			childs.emplace_back ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) );
			token = TreeFromStringLexer::next ( input );
		}

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		return ext::tree < DefaultSymbolType > ( std::move ( symbol ), std::move ( childs ) );
	}
}

} /* namespace tree */
