/*
 * arand.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <random>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Random generator binary", ' ', "0.01" );

		std::vector < std::string > allowed;
		allowed.push_back ( "FSM" );
		allowed.push_back ( "RFSM" );
		allowed.push_back ( "CFG" );
		allowed.push_back ( "RCFG" );
		allowed.push_back ( "RE" );
		allowed.push_back ( "ST" );
		allowed.push_back ( "SST" );
		allowed.push_back ( "UT" );
		allowed.push_back ( "UP" );
		allowed.push_back ( "RT" );
		allowed.push_back ( "RP" );
		allowed.push_back ( "RNP" );
		TCLAP::ValuesConstraint < std::string > allowedVals ( allowed );

		TCLAP::ValueArg < std::string > type ( "t", "type", "Type of generated structure", true, "FSM", & allowedVals );
		cmd.add ( type );
	
		TCLAP::ValueArg < unsigned > seed ( "s", "seed", "Seed random number", false, 0, "integer" );
		cmd.add ( seed );

		TCLAP::ValueArg < int > alphabetSize ( "", "terminals", "Number of terminals/alphabet size", false, 3, "integer" );
		cmd.add ( alphabetSize );

		TCLAP::SwitchArg integerSymbols ( "", "integerSymbols", "Use integer symbols instead of char symbols", false );
		cmd.add ( integerSymbols );

		TCLAP::SwitchArg randomizedAlphabet ( "", "randomizedTerminals", "Number of terminals/alphabet size", false );
		cmd.add ( randomizedAlphabet );

		TCLAP::ValueArg < double > density ( "", "density", "Automaton's/grammar's transition density", false, 10, "double" );
		cmd.add ( density );

		TCLAP::ValueArg < int > nonterminals ( "", "nonterminals", "Number of grammar's nononterminals", false, 5, "integer" );
		cmd.add ( nonterminals );

		TCLAP::ValueArg < int > nodes ( "", "nodes", "Number of automaton's or tree nodes", false, 6, "integer" );
		cmd.add ( nodes );

		TCLAP::ValueArg < int > states ( "", "states", "Number of automaton's states", false, 5, "integer" );
		cmd.add ( states );

		TCLAP::ValueArg < int > leaves ( "", "leaves", "Number of tree's/regexp's terminal nodes", false, 5, "integer" );
		cmd.add ( leaves );

		TCLAP::ValueArg < int > height ( "", "height", "Height of the tree/regexp", false, 3, "integer" );
		cmd.add ( height );

		TCLAP::ValueArg < int > length ( "", "length", "Length of the string", false, 5, "integer" );
		cmd.add ( length );

		TCLAP::ValueArg < int > rank ( "", "rank", "Maximal rank of tree nodes", false, 5, "integer" );
		cmd.add ( rank );

		TCLAP::SwitchArg singleNonlinearVariable ( "", "single_nonlinear_variable", "Generate single nonlinear variable", false );
		cmd.add ( singleNonlinearVariable );

		TCLAP::ValueArg < std::string > input ( "i", "input", "Input to use in generation", false, "-", "file" );
		cmd.add ( input );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		if ( seed.isSet ( ) )
			ext::random_devices::semirandom.seed ( seed.getValue ( ) );

		cli::Environment environment;
		environment.setBinding ( "stdout", "-" );

		environment.setBinding ( "alphabetSize", ext::to_string ( alphabetSize.getValue ( ) ) );

		if ( integerSymbols.isSet ( ) )
			environment.setBinding ( "integerSymbols", "true" );
		else
			environment.setBinding ( "integerSymbols", "false" );

		if ( randomizedAlphabet.isSet ( ) )
			environment.setBinding ( "randomizedAlphabet", "true" );
		else
			environment.setBinding ( "randomizedAlphabet", "false" );

		if ( ( density.getValue ( ) < 1 ) || ( density.getValue ( ) > 100 ) )
			throw exception::CommonException ( "You must specify density as a number between 1 and 100." );
		else
			environment.setBinding ( "density", ext::to_string ( density.getValue ( ) ) );

		environment.setBinding ( "nonterminals", ext::to_string ( nonterminals.getValue ( ) ) );

		environment.setBinding ( "nodes", ext::to_string ( nodes.getValue ( ) ) );

		environment.setBinding ( "states", ext::to_string ( states.getValue ( ) ) );

		environment.setBinding ( "leaves", ext::to_string ( leaves.getValue ( ) ) );

		environment.setBinding ( "height", ext::to_string ( height.getValue ( ) ) );

		environment.setBinding ( "length", ext::to_string ( length.getValue ( ) ) );

		environment.setBinding ( "rank", ext::to_string ( rank.getValue ( ) ) );

		if ( singleNonlinearVariable.isSet ( ) )
			environment.setBinding ( "singleNonlinearVariable", "true" );
		else
			environment.setBinding ( "singleNonlinearVariable", "false" );

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		if ( type.getValue ( ) == "RFSM"
		  || type.getValue ( ) == "RCFG"
		  || type.getValue ( ) == "SST" ) {
			std::string inputFile;
			if ( ! input.isSet ( ) )
				inputFile = "-";
			else
				inputFile = input.getValue ( );

			environment.setBinding ( "input", inputFile );
			cli::Parser parser ( cli::Lexer ( "execute <#input > $input" ) );
			parser.parse ( )->run ( environment );
		}

		measurements::end ( );
		measurements::start ( "Algorithm", measurements::Type::MAIN );

		std::string cliCommand;
		if ( type.getValue ( ) == "FSM" ) {
			cliCommand = "execute automaton::generate::RandomAutomatonFactory (int) #states (int) #alphabetSize (bool) #randomizedAlphabet (double) #density > $output";
		} else if ( type.getValue ( ) == "RFSM" ) {
			cliCommand = "execute automaton::generate::RandomizeAutomaton (int) #baseAutomaton > $output";
		} else if ( type.getValue ( ) == "CFG" ) {
			cliCommand = "execute grammar::generate::RandomGrammarFactory (int) #nonterminals (int) #alphabetSize (bool) #randomizedAlphabet (double) #density > $output";
		} else if ( type.getValue ( ) == "RCFG" ) {
			cliCommand = "execute grammar::generate::RandomizeGrammar (int) #baseGrammar > $output";
		} else if ( type.getValue ( ) == "RE" ) {
			cliCommand = "execute regexp::generate::RandomRegExpFactory (int) #leaves (int) #height (int) #alphabetSize (bool) #randomizedAlphabet > $output";
		} else if ( type.getValue ( ) == "ST" ) {
			cliCommand = "execute string::generate::RandomStringFactory (int) #length (int) #alphabetSize (bool) #randomizedAlphabet (bool) #integerSymbols > $output";
		} else if ( type.getValue ( ) == "SST" ) {
			cliCommand = "execute string::generate::RandomSubstringFactory (int) #length (int) #baseString > $output";
		} else if ( type.getValue ( ) == "UT" ) {
			cliCommand = "execute tree::generate::RandomUnrankedTreeFactory (int) #height (int) #nodes (int) #alphabetSize (bool) #randomizedAlphabet (int) #rank > $output";
		} else if ( type.getValue ( ) == "UP" ) {
			cliCommand = "execute tree::generate::RandomUnrankedPatternFactory (int) #height (int) #nodes (int) #alphabetSize (bool) #randomizedAlphabet (int) #rank > $output";
		} else if ( type.getValue ( ) == "RT" ) {
			cliCommand = "execute tree::generate::RandomRankedTreeFactory (int) #height (int) #nodes (int) #alphabetSize (bool) #randomizedAlphabet (int) #rank > $output";
		} else if ( type.getValue ( ) == "RP" ) {
			cliCommand = "execute tree::generate::RandomRankedPatternFactory (int) #height (int) #nodes (int) #alphabetSize (bool) #randomizedAlphabet (int) #rank > $output";
		} else if ( type.getValue ( ) == "RNP" ) {
			cliCommand = "execute tree::generate::RandomRankedNonlinearPatternFactory (int) #height (int) #nodes (int) #alphabetSize (bool) #randomizedAlphabet (bool) #singleNonlinearVariable (int) #rank > $output";
		} else {
			throw exception::CommonException ( "Invalid type." );
		}

		cli::Parser parser = cli::Parser ( cli::Lexer ( cliCommand ) );
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::start ( "Output write", measurements::Type::AUXILIARY );

		parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) common::Streams::measure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
